﻿using System;
using System.Linq;
using System.Collections.Generic;
using Eingabeverwaltung.Metamodell;

namespace Eingabeverwaltung.Parser
{
    /// <summary>
    /// This class holds Methods and Objects which are needed for more than one Listener.
    /// </summary>
    public class Listener_Helper
    {
        // helping objects for building
        public List<Sort> adt_sorts;
        public List<Operation> adt_ops;
        public List<Variable> adt_vars; // this is not always vars of adt, but in case of equation vars of equation
        public List<Variable> allQVars; // allQVars and fixedVars are just for saving their difference
        public List<Variable> fixedVars;
        public List<Axiom> adt_axioms;
        // Equation
        public Equation equation;
        // Tree
        public List<Tree> trees;
        public Listener_Helper(List<Sort> adt_sorts, List<Operation> adt_ops, List<Variable> adt_vars, List<Axiom> adt_axioms, Equation equation, List<Tree> trees)
        {
            this.adt_sorts = adt_sorts;
            this.adt_ops = adt_ops;
            this.adt_vars = adt_vars;
            this.adt_axioms = adt_axioms;
            this.equation = equation;
            this.trees = trees;
            allQVars = new List<Variable>();
            fixedVars = new List<Variable>();
    }
        // Equation
        public void EnterEq() 
        { 
            trees = new List<Tree>();
            allQVars = new List<Variable>();
            fixedVars = new List<Variable>();
        }
        public void ExitEq()
        {
            if (trees.Count != 2) throw new Exception("Calculated " + trees.Count + " Trees in Equation! The Trees are: " + TreeListToString(trees));
            Test_Equation_both_sides_have_same_Sort(trees[0], trees[1]);
            equation = new Equation(trees[0], trees[1], adt_vars);
        }

        public void ExitEqVarDef()
        {
            adt_vars = new List<Variable>(fixedVars);
            adt_vars.AddRange(allQVars);
        }

        public void EnterVars() => adt_vars = new List<Variable>();

        public void ExitVar(string variableName, string sortName)
        {
            Sort sort = findSortInADT_Sorts(sortName);
            Test_Variable_Sort_was_defined_in_ADT(sortName);
            Test_Variables_all_unique(variableName);
            adt_vars.Add(new Variable(variableName, sort));
        }

        // Trees
        /// <summary>
        /// TreeLeaf is whether a Variable or a constant (ComposedTree without children)
        /// </summary>
        /// <param name="treeName"></param>
        public void ExitTreeLeaf(string treeName)
        {
            Tree result = null;
            Variable variable = findVariableInADT_Vars(treeName);
            if (variable != null)
                result = new TreeVariable(variable);
            else
            {
                Operation operation = findOperationInADT_Ops(treeName, 0);
                if (operation != null)
                    result = new ComposedTree(operation, new List<Tree>());
            }
            if (result == null) throw new Exception("The Tree " + treeName + " could not be found in Vars or Operations of the ADT!");
            trees.Add(result);
        }

        public void ExitTreeNode(string treeName, int context_ChildCount, string context_Text)
        {
            int childCount = (int)Math.Ceiling(((double)(context_ChildCount - 3)) / 2); // -1 for BEZ, -1 for "(", -1 for ")", ceiling(/2) = -1 for each "," between 2 children
                                                                                        // All children of current tree are in this.trees because inner trees 1st, but sure is sure
            if (trees.Count >= childCount)
            {
                Operation operation = findOperationInADT_Ops(treeName, childCount);
                if (operation == null) throw new Exception("The Tree " + treeName + " could not be found in Operations of the ADT!");
                List<Tree> children = new List<Tree>();
                for (int i = 0; i < operation.sorts.Count - 1; i++)
                {
                    if (trees.Count == 0)
                        throw new Exception("Not enough created trees for the current tree! CurrentTreeName: " + treeName + "\n Context: " + context_Text + "\n added children: " + TreeListToString(children));
                    children.Add(trees.Last());
                    trees.RemoveAt(trees.Count - 1);
                }
                children.Reverse();
                Test_ComposedTree_Children_Sorts_are_correct(operation, children);
                Tree currentTree = new ComposedTree(operation, children);
                trees.Add(currentTree);
            }
            else
            {
                throw new Exception("Not all childen of the current Tree " + treeName + " are in trees!");
            }
        }

        // Helping functions

        /// <summary>
        /// Sets the context of variables for creating trees.
        /// Input: List of definition tables. Later tables are overwritten by earlier tables.
        /// One definition table is one List of Variables.
        /// </summary>
        /// <param name="contextList"></param>
        public List<Variable> getVariableContext(List<List<Variable>> contextList)
        {
            List<Variable> result = new List<Variable>();
            if (contextList.Count > 0) result.AddRange(contextList[0]);
            else return result; 
            for (int tableNum = 1; tableNum < contextList.Count; tableNum++)
            {
                List<Variable> table = contextList[tableNum]; // use table readOnly!
                foreach (Variable var in table)
                {
                    bool found = false;
                    foreach (Variable resVar in result)
                        if (var.name == resVar.name) { found = true; break; }
                    if (!found) result.Add(var);
                }
            }
            return result;
        }
        
        /// <summary>
        /// This Method is actually same as getVariableContext, but puts result to adt_vars.
        /// </summary>
        /// <param name="contextList"></param>
        public void setVariableContext(List<List<Variable>> contextList)
            => adt_vars = getVariableContext(contextList);

        /// <summary>
        /// Returns null if not able to find
        /// </summary>
        /// <param name="sortName"></param>
        /// <returns></returns>
        public Sort findSortInADT_Sorts(string sortName)
        {
            foreach (Sort sort in adt_sorts)
                if (sort.name == sortName)
                    return sort;
            return null;
        }

        /// <summary>
        /// Returns null if not able to find. Throws Exception if inputtedChildCount != expected ChildCount from the found Operation.
        /// </summary>
        /// <param name="operationName"></param>
        /// <param name="inputtedChildCount"></param>
        /// <returns></returns>
        public Operation findOperationInADT_Ops(string operationName, int inputtedChildCount) => findOperationInADT_Ops(operationName, inputtedChildCount, true);
        /// <summary>
        /// Returns null if not able to find
        /// </summary>
        /// <param name="operationName"></param>
        /// <param name="inputtedChildCount"></param>
        /// <param name="testChildCount"></param>
        /// <returns></returns>
        public Operation findOperationInADT_Ops(string operationName, int inputtedChildCount, bool testChildCount)
        {
            foreach (Operation operation in adt_ops)
            {
                if (operation.name == operationName)
                {
                    if (testChildCount) Test_ComposedTree_ChildCount_is_corrent(operation, inputtedChildCount);
                    return operation;
                }
            }
            return null;
        }

        /// <summary>
        /// Returns null if not able to find
        /// </summary>
        /// <param name="variableName"></param>
        /// <returns></returns>
        public Variable findVariableInADT_Vars(string variableName)
        {
            if (adt_vars == null) return null;
            foreach (Variable variable in adt_vars)
                if (variable.name == variableName) return variable;
            return null;

        }

        /// <summary>
        /// Returns null if not able to find
        /// </summary>
        /// <param name="variableName"></param>
        /// <returns></returns>
        public Variable findVariableInAllQVarsOrFixedVars(string variableName)
        {
            foreach (Variable variable in allQVars)
                if (variable.name == variableName) return variable;
            foreach (Variable variable in fixedVars)
                if (variable.name == variableName) return variable;
            return null;
        }

        /// <summary>
        /// Returns null if not able to find
        /// Finds Variable in specified Variable list
        /// </summary>
        /// <param name="variableName"></param>
        /// <returns></returns>
        public Variable findVariable(string variableName, List<Variable> variables)
        {
            if (variables == null) return null;
            foreach (Variable variable in variables)
                if (variable.name == variableName) return variable;
            return null;
        }

        /// <summary>
        /// Returns null if not able to find
        /// </summary>
        /// <param name="axiomName"></param>
        /// <returns></returns>
        public Axiom findAxiomInADT_Axioms(string axiomName)
        {
            foreach (Axiom axiom in adt_axioms)
                if (axiom.name == axiomName) return axiom;
            return null;
        }

        public string TreeListToString(List<Tree> treeList)
        {
            string result = "";
            foreach (Tree tree in treeList)
                result += tree.ToString() + ',';
            if (result.Length > 0) result.Remove(result.Length - 1);
            return result;
        }

        // Alle Offiziell notwendige Typtests

        /// <summary>
        /// This Test get always activeted when a Tree gets parsed.
        /// </summary>
        /// <param name="operation"></param>
        /// <param name="inputtedChildCount"></param>
        private void Test_ComposedTree_ChildCount_is_corrent(Operation operation, int inputtedChildCount)
        {
            if (operation.sorts.Count != inputtedChildCount + 1)
                throw new Exception("The Operation_Tree " + operation.name + " was inputted with " + inputtedChildCount + " children, expected " + (operation.sorts.Count - 1) + " children!");
        }
        private void Test_ComposedTree_Children_Sorts_are_correct(Operation operation, List<Tree> children)
        {
            for (int i = 0; i < children.Count; i++)
            {
                string type = "";
                if (children[i] as ComposedTree != null) type = "ComposedTree";
                else type = "Variable";
                if (children[i].sort != operation.sorts[i])
                    throw new Exception("Error while building the tree " + operation.name + ": The the child (" + type + ") " + children[i].name
                                        + " with the Sort " + children[i].sort.name + " is not " + operation.sorts[i].name + "!");
            }
        }
        /// <summary>
        /// This Test is done by getting the Variable or Operation from the TreeName in the big if clause of ExitMyTree method.
        /// Moving that Test to this Method is not possible.
        /// Implementing that Test here again is nonsense.
        /// This Method does actually nothing, its just a dummy for knowing, that this Test is implemented.
        /// </summary>
        private void Test_Tree_Variables_were_defined_in_ADT() {/* Nothing here need to be. */}
        private void Test_Equation_both_sides_have_same_Sort(Tree leftTree, Tree rightTree)
        {
            if (leftTree.sort.name != rightTree.sort.name)
                throw new Exception("The Equation has not the same Sort on its 2 sides! Sort of leftTree: " + leftTree.sort.name + " Sort of rightTree: " + rightTree.sort.name);
        }

        /// <summary>
        /// This Test is activated always when the Sort of a Variable gets read from the parser,
        /// so as result the Sort of every Variable was defined in the ADT.
        /// </summary>
        /// <param name="sortName"></param>
        private void Test_Variable_Sort_was_defined_in_ADT(string sortName)
        {
            if (findSortInADT_Sorts(sortName) == null) throw new Exception("Found a Sort " + sortName + " in Variable that was not defined in Sorts!");
        }

        /// <summary>
        /// This Test is activated always when a new Variable gets added to the variableList of the ADT / Equation,
        /// so as result all Variables are unique.
        /// </summary>
        /// <param name="variableName"></param>
        private void Test_Variables_all_unique(string variableName)
        {
            if (findVariableInADT_Vars(variableName) != null
                || findVariableInAllQVarsOrFixedVars(variableName) != null
                ) throw new Exception("The Variable " + variableName + " was defined more than 1 times.!");
        }
    }
}
